#----------------------------------------------------------------------------
#
#   MatSheet - Sheet View
#
#----------------------------------------------------------------------------

from GUI import ScrollableView, TextField, rgb
from GUI.Geometry import inset_rect
from GUI.StdFonts import application_font
from GUI.StdColors import black, yellow, \
    selection_forecolor, selection_backcolor
from sheetdoc import BCCoords, BCRect

def gray(x):
    return rgb(x, x, x)

class SheetView(ScrollableView):
    #  model         SheetDoc
    #  data_font     Font
    #  cell_editor   CellEditor
    #  editing_cell  BCCoords      Cell being edited
    #  clicked_cell  BCCoords      Cell clicked on to start drag

    block_border_size = 2
    cell_border_size = 1
    data_color = black
    faded_data_color = gray(0.75)
    cell_border_color = gray(0.75)
    #off_sheet_color = gray(0.5)
    cell_top_margin = 4
    cell_bottom_margin = 2
    edit_layout_border_width = 5
    
    editing_cell = None
    clicked_cell = None
    
    def __init__(self, **kwds):
        ScrollableView.__init__(self, **kwds)
        self.cell_editor = CellEditor() #border = False)
        self.data_font = self.cell_editor.font
        self.update_extent()
    
    def setup_menus(self, m):
        sheet = self.model
        m.edit_layout_cmd.enabled = True
        m.edit_layout_cmd.checked = self.model.editing_layout
    
    def edit_layout_cmd(self):
        self.close_cell()
        sheet = self.model
        sheet.set_selection(None)
        sheet.toggle_editing_layout()
    
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def draw(self, c, ur):
        c.erase_rect(ur)
        sheet = self.model
        gap = self.block_border_size
        y = 0
        ygap = 0
        editing_layout = sheet.editing_layout
        selection = sheet.selection
        for row in xrange(sheet.num_rows):
            bh = self.block_height(sheet.row_headers[row])
            x = 0
            xgap = 0
            for col in xrange(sheet.num_columns):
                header = sheet.column_headers[col]
                bw = self.block_width(header)
                br = (x, y, x + bw, y + bh)
                if editing_layout:
                    if selection and BCCoords((row, col), (0, 0)) in selection:
                        c.fillcolor = selection_backcolor
                        c.fill_rect(br)
                    self.draw_edit_layout_border(c, br)
                self.draw_block(c, br, row, col)
                x += xgap + bw
                xgap = gap
            y += ygap + bh
            ygap = gap
    
    def draw_edit_layout_border(self, c, rect):
        d = self.edit_layout_border_width
        c.gsave()
        c.newpath()
        c.rect(rect)
        c.rect(inset_rect(rect, (d, d)))
        c.clip()
        l, t, r, b = rect
        x = l
        y = t
        w = r - l
        h = b - t
        i = 0
        colors = (black, yellow)
        xmax = r + h
        while x < xmax:
            c.fillcolor = colors[i]
            c.fill_poly([(x, y), (x+d, y), (x+d-h, y+h), (x-h, y+h)])
            i = 1 - i
            x += d
        c.grestore()
    
    def draw_block(self, c, rect, brow, bcol):
        sheet = self.model
        col_header = sheet.column_headers[bcol]
        block = sheet.blocks[brow][bcol]
        l, t, r, b = rect
        format = "%s"
        font = self.data_font
        c.font = font
        cell_width = col_header.cell_width
        cell_height = self.cell_height()
        base = font.ascent + self.cell_top_margin
        gap = self.cell_border_size
        nrows = block.num_rows
        ncols = block.num_columns
        data = block.data
        align = block.align
        lm = block.left_margin
        rm = block.right_margin
        editing_layout = self.model.editing_layout
        if editing_layout:
            color = self.faded_data_color
        else:
            color = self.data_color
        c.textcolor = color
        selection = sheet.selection
        y = t
        ygap = 0
        for row in xrange(nrows):
            x = l
            xgap = 0
            for col in xrange(ncols):
                cell_selected = False
                if selection and not editing_layout:
                    bc = BCCoords((brow, bcol), (row, col))
                    if bc in selection:
                        cell_selected = True
                        c.fillcolor = selection_backcolor
                        c.fill_rect(self.cell_rect(bc))
                if cell_selected:
                    c.textcolor = selection_forecolor
                else:
                    c.textcolor = color
                text = format % data[row][col]
                yb = y + base
                if align == 'l':
                    xb = x + lm
                else:
                    xb = x + cell_width - rm - font.width(text)
                c.moveto(xb, yb)
                c.show_text(text)
                x += xgap + cell_width
                xgap = gap
            y += ygap + cell_height
            ygap = gap
        c.fillcolor = self.cell_border_color
        for row in xrange(1, nrows):
            y = t + row * (cell_height + gap)
            c.fill_rect((l, y - gap, r, y))
        for col in xrange(1, ncols):
            x = l + col * (cell_width + gap)
            c.fill_rect((x - gap, t, x, b))

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def update_extent(self):
        sheet = self.model
        gap = self.block_border_size
        w = (sheet.num_columns - 1) * gap
        h = (sheet.num_rows - 1) * gap
        for hdr in sheet.column_headers:
            w += self.block_width(hdr)
        for hdr in sheet.row_headers:
            h += self.block_height(hdr)
        self.extent = (w, h)

    def mouse_down(self, e):
        self.close_cell()
        bc = self.find_cell(e.position)
        if not e.shift:
            self.clicked_cell = bc
        self.mouse_drag(e)
    
    def mouse_drag(self, e):
        self.mouse_drag_or_up(e)
    
    def mouse_up(self, e):
        self.mouse_drag_or_up(e)
    
    def mouse_drag_or_up(self, e):
        bc0 = self.clicked_cell
        bc1 = self.find_cell(e.position)
        if bc0 and bc1:
            sheet = self.model
            sheet.set_selection(BCRect(bc1, self.clicked_cell))
            if e.kind == 'mouse_up' and bc0 == bc1:
                if e.shift:
                    sheet.set_selection(None)
                    self.clicked_cell = None
                else:
                    if not sheet.editing_layout:
                        self.open_cell(bc1)
    
    def open_cell(self, bc):
        #  Start editing specified cell
        self.close_cell()
        sheet = self.model
        sheet.selection = BCRect(bc, bc)
        self.editing_cell = bc
        field = self.cell_editor
        field.bounds = self.cell_rect(bc)
        field.text = str(sheet.blocks[bc.brow][bc.bcol].data[bc.crow, bc.ccol])
        self.add(field)
        field.become_target()
        self.invalidate()
        
    def close_cell(self):
        #  Stop editing cell, if any
        self.model.sync()
    
    def move_up(self):
        self.move_to_adjacent_cell(-1, 0)

    def move_down(self):
        self.move_to_adjacent_cell(1, 0)

    def move_left(self):
        self.move_to_adjacent_cell(0, -1)
    
    def move_right(self):
        self.move_to_adjacent_cell(0, 1)
    
    def move_to_adjacent_cell(self, drow, dcol):
        bc = self.editing_cell
        if bc:
            self.close_cell()
            brow = bc.brow
            bcol = bc.bcol
            crow = bc.crow
            ccol = bc.ccol
            block = self.model.blocks[brow][bcol]
            nrows, ncols = block.data.shape
            crow += drow
            ccol += dcol
            if crow >= nrows:
                crow = 0
                ccol += 1
                if ccol >= ncols:
                    ccol = 0
            elif crow < 0:
                crow = nrows - 1
                ccol -= 1
                if ccol < 0:
                    ccol = ncols - 1
            if ccol >= ncols:
                ccol = 0
                crow += 1
                if crow >= nrows:
                    crow = 0
            elif ccol < 0:
                ccol = ncols - 1
                crow -= 1
                if crow < 0:
                    crow = nrows - 1
            bc2 = BCCoords((brow, bcol), (crow, ccol))
            self.open_cell(bc2)
    
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    def model_sync(self, model):
        #  Broadcast by model to force closing of any open cells
        bc = self.editing_cell
        if bc:
            text = self.cell_editor.text
            sheet = self.model
            value = sheet.blocks[bc.brow][bc.bcol].text_to_value(text)
            if self.window.target is self.cell_editor:
                self.become_target()
            self.remove(self.cell_editor)
            self.editing_cell = None
            sheet.do_change_cell(bc, value)
    
    def find_cell(self, p):
        #  Returns block and cell coordinates of the cell containing
        #  point p, or None. If editing layout, returns coordinates of
        #  the top left cell of the clicked block.
        px, py = p
        x0 = y0 = 0
        for brow, y0, y1 in self.block_row_spans():
            if y0 <= py < y1:
                break
        else:
            return
        for bcol, x0, x1 in self.block_column_spans():
            if x0 <= px < x1:
                break
        else:
            return
        sheet = self.model
        if sheet.editing_layout:
            crow = ccol = 0
        else:
            gap = self.cell_border_size
            crow = int((py - y0) // (self.cell_height() + gap))
            ccol = int((px - x0) // (sheet.column_headers[bcol].cell_width + gap))
        return BCCoords((brow, bcol), (crow, ccol))
    
    def block_row_spans(self):
        #  Generates sequence of (row_num, top, bottom)
        sheet = self.model
        hdrs = sheet.row_headers
        gap = self.block_border_size
        y0 = 0
        ch = self.cell_height() + self.cell_border_size
        for i, hdr in enumerate(hdrs):
            y1 = y0 + ch * hdr.num_cells
            yield i, y0, y1
            y0 = y1
        
    def block_column_spans(self):
        #  Generates sequence of (col_num, left, right)
        sheet = self.model
        hdrs = sheet.column_headers
        bgap = self.block_border_size
        cgap = self.cell_border_size
        x0 = 0
        for i, hdr in enumerate(hdrs):
            x1 = x0 + (hdr.cell_width + cgap) * hdr.num_cells
            yield i, x0, x1
            x0 = x1

    def block_width(self, header):
        #  Returns width of block column excluding block borders
        n = header.num_cells
        return n * header.cell_width + (n - 1) * self.cell_border_size
    
    def block_height(self, header):
        #  Returns height of block row excluding block borders
        n = header.num_cells
        return n * self.cell_height() + (n - 1) * self.cell_border_size
    
    def cell_height(self):
        #  Returns height of a cell excluding cell borders
        tm = self.cell_top_margin
        bm = self.cell_bottom_margin
        return self.data_font.line_height + tm + bm

    def cell_rect(self, bc):
        #  Returns rectangle covered by a cell excluding cell borders
        x0, y0 = self.block_topleft(bc.brow, bc.bcol)
        cw = self.model.column_headers[bc.bcol].cell_width
        ch = self.cell_height()
        gap = self.cell_border_size
        l = x0 + bc.ccol * (cw + gap)
        t = y0 + bc.crow * (ch + gap)
        return (l, t, l + cw, t + ch)
    
    def block_topleft(self, brow, bcol):
        #  Returns position of top left corner of block
        x = y = 0
        gap = self.block_border_size
        sheet = self.model
        for i in xrange(brow):
            y += self.block_height(sheet.row_headers[i]) + gap
        for i in xrange(bcol):
            x += self.block_width(sheet.column_headers[i]) + gap
        return (x, y)

#----------------------------------------------------------------------------

class CellEditor(TextField):

    def key_down(self, e):
        c = e.char
        if c == '\r':
            if e.key == 'enter':
                self.container.close_cell()
            else:
                if e.shift:
                    self.container.move_up()
                else:
                    self.container.move_down()
        elif c == '\t':
            if e.shift:
                self.container.move_left()
            else:
                self.container.move_right()
        else:
            TextField.key_down(self, e)

    def untargeted(self):
        self.container.close_cell()
